/*
 *  linux/drivers/ide/pdc202xx.c	Version 0.36	Aug. 01, 2002
 *
 *  Copyright (C) 1998-2000	Andre Hedrick <andre@linux-ide.org>
 *  May be copied or modified under the terms of the GNU General Public License
 *
 *  Copyright (C) 1999-2002 Promise Technology, Inc.
 *  Author: Frank Tiernan <frankt@promise.com>
 *          Hank Yang <support@promise.com.tw>
 *  Released under terms of General Public License
 * 
 *  See linux/MAINTAINERS for address of current maintainer.
 *
 *  This file provides support for the advanced features and bugs
 *  of IDE interfaces using the Promise Ultra Series IDE interface chip.
 *  
 *  Promise Ultra33 cards with BIOS v1.20 through 1.28 will need this
 *  compiled into the kernel if you have more than one card installed.
 *  Note that BIOS v1.29 is reported to fix the problem.  Since this is
 *  safe chipset tuning, including this support is harmless
 *
 *  Promise Ultra66 cards with BIOS v1.11 this
 *  compiled into the kernel if you have more than one card installed.
 *
 *  Promise Ultra100 cards with BIOS v2.01 this
 *  compiled into the kernel if you have more than one card installed.
 *
 *  Promise Ultra100TX2 with BIOS v2.10 & Ultra133TX2 with BIOS v2.20
 *  support 8 hard drives on UDMA mode.
 *
 *  Linux kernel was incorrectly detects FastTrak ATA-RAID controllers
 *  as Ultra IDE Controllers.
 *  However, UNLESS check CLASS ID when ide_scan_pcidev.
 *  That's you can use FastTrak ATA-RAID controllers.
 *
 *  History :
 *  05/22/01    v1.20 b1
 *           (1) support PDC20268 controller
 *           (2) fix cable judge function (new_chip only)
 *  08/22/01    v1.20 b2
 *           (1) support ATA-133 PDC20269/75 controller
 *           (2) support UDMA Mode 6
 *           (3) fix proc report information
 *           (4) set ATA133 timing
 *           (5) fix ultra dma bit 14 selectable
 *           (6) support 32bit LBA
 *  09/11/01    v1.20 b3 
 *           (1) fix eighty_ninty_three()
 *           (2) fix offset address 0x1c~0x1f (old_chip only)
 *  10/30/01    v1.20 b4
 *           (1) fix 48bit LBA HOB bit
 *           (2) support HDIO_SCAN_HWIF Ioctl under PIO mode if necessary
 *  11/02/01    v1.20.0.5
 *           (1) could be patched with ext3 filesystem code
 *  11/06/01    v1.20.0.6
 *           (1) fix LBA48 drive running without Promise controllers
 *           (2) fix LBA48 drive running under PIO modes
 *  01/28/02    v1.20.0.6
 *           (1) release for linux IDE Group kernel 2.4.18
 *  07/11/02    v1.20.0.7
 *           (1) add version and controller info to proc
 *           (2) fix PDC20262 running with 48bit [disable]
 *           (3) Add quirk drive lists for PDC20265/67
 *  07/16/02	v1.20.0.8
 *           (1) Fix adj value and speed_rate in pdc202xx_tune_chipset
 *           (2) Fix indexreg/datareg while set ATA133 timing
 *           (3) Update pdc202xx_reset function
 *           (4) Update pci_init_pdc202xx function
 *           (5) Add pdc202xx_marvell_idle function 
 *               for Marvell bridge's Idle Immediate command
 *  08/01/02	v1.20.0.9
 *           (1) Fix IO-ports for TX2 family. (BA#4)
 *           (2) Update quirk_drives string.
 *           (3) Rename new_chip to tx2_family.
 *           (4) Update pdc202xx proc info for two adapters
 *           (5) Recheck down mode if using ata33 cable
 *
 */
#define VERSION	"1.20.0.9"
#define VERDATE "2002-08-01"

#include <linux/config.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/delay.h>
#include <linux/timer.h>
#include <linux/mm.h>
#include <linux/ioport.h>
#include <linux/blkdev.h>
#include <linux/hdreg.h>
#include <linux/interrupt.h>
#include <linux/pci.h>
#include <linux/init.h>
#include <linux/ide.h>

#include <asm/io.h>
#include <asm/irq.h>

#include "ide_modes.h"

#define PDC202XX_DEBUG_DRIVE_INFO		0
#define PDC202XX_DECODE_REGISTER_INFO		0

#define DISPLAY_PDC202XX_TIMINGS

#ifndef SPLIT_BYTE
#define SPLIT_BYTE(B,H,L)	((H)=(B>>4), (L)=(B-((B>>4)<<4)))
#endif

#if defined(DISPLAY_PDC202XX_TIMINGS) && defined(CONFIG_PROC_FS)
#include <linux/stat.h>
#include <linux/proc_fs.h>

static int pdc202xx_get_info(char *, char **, off_t, int);
extern int (*pdc202xx_display_info)(char *, char **, off_t, int); /* ide-proc.c */
byte pdc202xx_proc = 0;

#define PDC202_MAX_DEVS		5

static struct pci_dev *pdc202_devs[PDC202_MAX_DEVS];
static int n_pdc202_devs;
static struct hd_driveid *id[MAX_DRIVES * 4]; /* 4 channels proc info */

const char *pdc_quirk_drives[] = {
	"QUANTUM FIREBALLlct08 08",
	"QUANTUM FIREBALLP KA",
	"QUANTUM FIREBALLP LM",
	"QUANTUM FIREBALLP KX",
	NULL
};

static char * pdc202xx_info (char *buf, struct pci_dev *dev)
{
	char *p = buf;

	u32 bibma	= pci_resource_start(dev, 4);
	u32 index_p	= bibma + 0x01;
	u32 data_p	= index_p + 2;
	u32 index_s	= bibma + 0x09;
	u32 data_s	= index_s + 2;
	u16 reg50h 	= 0, word88 = 0;
	unsigned short mask_p 	= (1<<10);
	unsigned short mask_s 	= (1<<11);
	unsigned int cable_p = 1, cable_s = 1;	/* Set default to 40-pin */
	static int udmasel[4 * MAX_DRIVES], piosel[4 * MAX_DRIVES];
	unsigned int i = 0, hd = 0, h, k = 0;
	ide_hwif_t *hwif;
	unsigned long base = pci_resource_start(dev, 0);

        /*
         * at that point bibma+0x2 et bibma+0xa are byte registers
         * to investigate:
         */
	u8 c0	= inb_p((unsigned short)bibma + 0x02);
	u8 c1	= inb_p((unsigned short)bibma + 0x0a);
	
	for (h = 2; h < MAX_HWIFS; ++h) {
		hwif = &ide_hwifs[h];
		if (hwif->io_ports[IDE_DATA_OFFSET] == base)
			if (hwif->chipset == ide_pci)
				k = ((hwif->index > 3) & (hwif->index < 6)) ? 4 : 0 ;
	}

	switch(dev->device) {
		case PCI_DEVICE_ID_PROMISE_20267:
		case PCI_DEVICE_ID_PROMISE_20265:
		case PCI_DEVICE_ID_PROMISE_20262:
		case PCI_DEVICE_ID_PROMISE_20246:
			pci_read_config_word(dev, 0x50, &reg50h);
			cable_p = (reg50h & mask_p);
			cable_s = (reg50h & mask_s);
			break;
		default:
			OUT_BYTE(0x0b, index_p);
			cable_p = (IN_BYTE(data_p) & 0x04);
			OUT_BYTE(0x0b, index_s);
			cable_s = (IN_BYTE(data_s) & 0x04);
	}

	p += sprintf(p, "\nPROMISE Ultra series driver Ver %s \t Date: %s \n",
		       	VERSION, VERDATE);
	p += sprintf(p, "Adapter : ");
	switch(dev->device) {
		case PCI_DEVICE_ID_PROMISE_20275:
			p += sprintf(p, "MBUltra133\n"); break;
		case PCI_DEVICE_ID_PROMISE_20269:
			p += sprintf(p, "Ultra133 TX2\n"); break;
		case PCI_DEVICE_ID_PROMISE_20268:
			p += sprintf(p, "Ultra100 TX2\n"); break;
		case PCI_DEVICE_ID_PROMISE_20267:
			p += sprintf(p, "Ultra100\n"); break;
		case PCI_DEVICE_ID_PROMISE_20265:
			p += sprintf(p, "Ultra100 on M/B\n"); break;
		case PCI_DEVICE_ID_PROMISE_20262:
			p += sprintf(p, "Ultra66\n"); break;
		case PCI_DEVICE_ID_PROMISE_20246:
			p += sprintf(p, "Ultra33\n");
			reg50h |= 0x0c00; break;
		case PCI_DEVICE_ID_PROMISE_20277:
			p += sprintf(p, "SBFastTrak 133 Lite\n"); break;
		case PCI_DEVICE_ID_PROMISE_20276:
			p += sprintf(p, "MBFastTrak 133 Lite\n"); break;
		case PCI_DEVICE_ID_PROMISE_20271:
			p += sprintf(p, "FastTrak TX2000\n"); break;
		case PCI_DEVICE_ID_PROMISE_20270:
			p += sprintf(p, "FastTrak LP/TX2/TX4\n"); break;
		default:
			p += sprintf(p, "Ultra Series\n"); break;
	}
	p += sprintf(p, "VendorID: %#x  DeviceID: %#x  ClassID: 0x%4.4x  IRQ: %#x\n",
		dev->vendor, dev->device, dev->class >> 8, dev->irq );
	p += sprintf(p, "--------------- Primary Channel ---------------- Secondary Channel -------------\n");
	p += sprintf(p, "Status: \t %s \t \t \t %s \n",
		(c0 & 0x80) ? "disabled" : "enabled ",
		(c1 & 0x80) ? "disabled" : "enabled ");
	p += sprintf(p, "Cable type: \t %s \t \t \t %s \n",
		cable_p ? "40-pin" : "80-pin",
		cable_s ? "40-pin" : "80-pin");
	p += sprintf(p, "--------------- drive0 --------- drive1 -------- drive0 -------- drive1 --------\n");
	p += sprintf(p, "DMA enabled: \t %s \t \t %s \t \t %s \t \t %s\n",
		(id[0+k]!=NULL && (c0 & 0x20)) ? "yes" : "no",
		(id[1+k]!=NULL && (c0 & 0x40)) ? "yes" : "no",
		(id[2+k]!=NULL && (c1 & 0x20)) ? "yes" : "no",
		(id[3+k]!=NULL && (c1 & 0x40)) ? "yes" : "no");
	p += sprintf(p, "LBA48 Mode: \t %s \t \t %s \t \t %s \t \t %s\n",
		(id[0+k]!=NULL && (id[0+k]->command_set_2 & 0x0400)) ? "yes" : "no",
		(id[1+k]!=NULL && (id[1+k]->command_set_2 & 0x0400)) ? "yes" : "no",
		(id[2+k]!=NULL && (id[2+k]->command_set_2 & 0x0400)) ? "yes" : "no",
		(id[3+k]!=NULL && (id[3+k]->command_set_2 & 0x0400)) ? "yes" : "no");

	for(hd=0; hd<4; hd++) {
		if (id[ hd+k ] == NULL)
			continue;
		word88 = id[ hd+k ]->dma_ultra;
		for (i=7; i>=0; i--)
			if (word88 >> (i+8)) {
				udmasel[ hd+k ] = i;
				/* get select UDMA mode */
				break;
			}
		piosel[ hd+k ]=(id[ hd+k ]->eide_pio_modes >= 0x02) ? 4 : 3;
        }
	p += sprintf(p, "UDMA Mode: \t %d \t \t %d \t \t %d \t \t %d\n",
		udmasel[0+k], udmasel[1+k],
	       	udmasel[2+k], udmasel[3+k]);
	p += sprintf(p, "PIO Mode: \t %d \t \t %d \t \t %d \t \t %d\n",
		piosel[0+k], piosel[1+k],
	       	piosel[2+k], piosel[3+k]);
	p += sprintf(p, "--------------------------------------------------------------------------------\n");
	return (char *)p;
}

static int pdc202xx_get_info (char *buffer, char **addr, off_t offset, int count)
{
	char *p = buffer;
	int i;
	
	for (i = 0; i < n_pdc202_devs; i++) {
		struct pci_dev *dev	= pdc202_devs[i];
		p = pdc202xx_info(p, dev);
	}
	return p-buffer;	/* => must be less than 4k! */
}
#endif  /* defined(DISPLAY_PDC202XX_TIMINGS) && defined(CONFIG_PROC_FS) */

/* A Register */
#define	SYNC_ERRDY_EN	0xC0

#define	SYNC_IN		0x80	/* control bit, different for master vs. slave drives */
#define	ERRDY_EN	0x40	/* control bit, different for master vs. slave drives */
#define	IORDY_EN	0x20	/* PIO: IOREADY */
#define	PREFETCH_EN	0x10	/* PIO: PREFETCH */

#define	PA3		0x08	/* PIO"A" timing */
#define	PA2		0x04	/* PIO"A" timing */
#define	PA1		0x02	/* PIO"A" timing */
#define	PA0		0x01	/* PIO"A" timing */

/* B Register */

#define	MB2		0x80	/* DMA"B" timing */
#define	MB1		0x40	/* DMA"B" timing */
#define	MB0		0x20	/* DMA"B" timing */

#define	PB4		0x10	/* PIO_FORCE 1:0 */

#define	PB3		0x08	/* PIO"B" timing */	/* PIO flow Control mode */
#define	PB2		0x04	/* PIO"B" timing */	/* PIO 4 */
#define	PB1		0x02	/* PIO"B" timing */	/* PIO 3 half */
#define	PB0		0x01	/* PIO"B" timing */	/* PIO 3 other half */

/* C Register */
#define	IORDYp_NO_SPEED	0x4F
#define	SPEED_DIS	0x0F

#define	DMARQp		0x80
#define	IORDYp		0x40
#define	DMAR_EN		0x20
#define	DMAW_EN		0x10

#define	MC3		0x08	/* DMA"C" timing */
#define	MC2		0x04	/* DMA"C" timing */
#define	MC1		0x02	/* DMA"C" timing */
#define	MC0		0x01	/* DMA"C" timing */

#if PDC202XX_DECODE_REGISTER_INFO

#define REG_A		0x01
#define REG_B		0x02
#define REG_C		0x04
#define REG_D		0x08

static void decode_registers (byte registers, byte value)
{
	byte	bit = 0, bit1 = 0, bit2 = 0;

	switch(registers) {
		case REG_A:
			bit2 = 0;
			printk("A Register ");
			if (value & 0x80) printk("SYNC_IN ");
			if (value & 0x40) printk("ERRDY_EN ");
			if (value & 0x20) printk("IORDY_EN ");
			if (value & 0x10) printk("PREFETCH_EN ");
			if (value & 0x08) { printk("PA3 ");bit2 |= 0x08; }
			if (value & 0x04) { printk("PA2 ");bit2 |= 0x04; }
			if (value & 0x02) { printk("PA1 ");bit2 |= 0x02; }
			if (value & 0x01) { printk("PA0 ");bit2 |= 0x01; }
			printk("PIO(A) = %d ", bit2);
			break;
		case REG_B:
			bit1 = 0;bit2 = 0;
			printk("B Register ");
			if (value & 0x80) { printk("MB2 ");bit1 |= 0x80; }
			if (value & 0x40) { printk("MB1 ");bit1 |= 0x40; }
			if (value & 0x20) { printk("MB0 ");bit1 |= 0x20; }
			printk("DMA(B) = %d ", bit1 >> 5);
			if (value & 0x10) printk("PIO_FORCED/PB4 ");
			if (value & 0x08) { printk("PB3 ");bit2 |= 0x08; }
			if (value & 0x04) { printk("PB2 ");bit2 |= 0x04; }
			if (value & 0x02) { printk("PB1 ");bit2 |= 0x02; }
			if (value & 0x01) { printk("PB0 ");bit2 |= 0x01; }
			printk("PIO(B) = %d ", bit2);
			break;
		case REG_C:
			bit2 = 0;
			printk("C Register ");
			if (value & 0x80) printk("DMARQp ");
			if (value & 0x40) printk("IORDYp ");
			if (value & 0x20) printk("DMAR_EN ");
			if (value & 0x10) printk("DMAW_EN ");

			if (value & 0x08) { printk("MC3 ");bit2 |= 0x08; }
			if (value & 0x04) { printk("MC2 ");bit2 |= 0x04; }
			if (value & 0x02) { printk("MC1 ");bit2 |= 0x02; }
			if (value & 0x01) { printk("MC0 ");bit2 |= 0x01; }
			printk("DMA(C) = %d ", bit2);
			break;
		case REG_D:
			printk("D Register ");
			break;
		default:
			return;
	}
	printk("\n        %s ", (registers & REG_D) ? "DP" :
				(registers & REG_C) ? "CP" :
				(registers & REG_B) ? "BP" :
				(registers & REG_A) ? "AP" : "ERROR");
	for (bit=128;bit>0;bit/=2)
		printk("%s", (value & bit) ? "1" : "0");
	printk("\n");
}

#endif /* PDC202XX_DECODE_REGISTER_INFO */

#if 0
static byte pdc202xx_ratemask (ide_drive_t *drive)
{
	struct pci_dev *dev = HWIF(drive)->pci_dev;
	byte mode = 0x00;

	switch(dev->device) {
		case PCI_DEVICE_ID_PROMISE_20277:
		case PCI_DEVICE_ID_PROMISE_20276:
		case PCI_DEVICE_ID_PROMISE_20275:
		case PCI_DEVICE_ID_PROMISE_20271:
		case PCI_DEVICE_ID_PROMISE_20269:
			{ mode |= 0x04; break; }
		case PCI_DEVICE_ID_PROMISE_20270:
		case PCI_DEVICE_ID_PROMISE_20268:
		case PCI_DEVICE_ID_PROMISE_20267:
		case PCI_DEVICE_ID_PROMISE_20265:
			{ mode |= 0x03; break; }
		case PCI_DEVICE_ID_PROMISE_20263:
		case PCI_DEVICE_ID_PROMISE_20262:
			{ mode |= 0x02; break; }
		case PCI_DEVICE_ID_PROMISE_20246:
			{ mode |= 0x01; break; }
		default:
			return (mode &= ~0xF8);
	}

	if (!eighty_ninty_three(drive)) {
		mode &= ~0xFE;
		mode |= 0x01;
	}
	return (mode &= ~0xF8);
}

static byte pdc202xx_ratefilter (ide_drive_t *drive, byte speed)
{
#ifdef CONFIG_BLK_DEV_IDEDMA
	byte mode = pdc202xx_ratemask(drive);

	switch(mode) {
		case 0x04:	while (speed > XFER_UDMA_6) speed--; break;
		case 0x03:	while (speed > XFER_UDMA_5) speed--; break;
		case 0x02:	while (speed > XFER_UDMA_4) speed--; break;
		case 0x01:	while (speed > XFER_UDMA_2) speed--; break;
		case 0x00:
		default:	while (speed > XFER_MW_DMA_2) speed--; break;
			break;
	}
#else
	while (speed > XFER_PIO_4) speed--;
#endif /* CONFIG_BLK_DEV_IDEDMA */
//	printk("%s: mode == %02x speed == %02x\n", drive->name, mode, speed);
	return speed;
}

#else
static byte pdc202xx_ratefilter (ide_drive_t *drive, byte speed)
{
	return speed;
}
#endif

static int check_in_drive_lists (ide_drive_t *drive, const char **list)
{
	struct hd_driveid *id = drive->id;

	if (pdc_quirk_drives == list) {
		while (*list) {
			if (strstr(id->model, *list++)) {
				return 2;
			}
		}
	} else {
		while (*list) {
			if (!strcmp(*list++,id->model)) {
				return 1;
			}
		}
	}
	return 0;
}

static int pdc202xx_tune_chipset (ide_drive_t *drive, byte xferspeed)
{
	ide_hwif_t *hwif	= HWIF(drive);
	struct pci_dev *dev	= hwif->pci_dev;
	byte speed		= pdc202xx_ratefilter(drive, xferspeed);

	unsigned int		drive_conf;
	byte			drive_pci, AP, BP, CP, DP;
	byte			TA = 0, TB = 0, TC = 0;
	unsigned int 		err, i = 0, j = hwif->channel ? 2 : 0;
	unsigned int k 		= ((hwif->index > 3) & (hwif->index < 6)) ? 4 : 0 ;

	drive_pci = 0x60 + (drive->dn << 2);

	if ((drive->media != ide_disk) && (speed < XFER_SW_DMA_0))
		return -1;

	pci_read_config_dword(dev, drive_pci, &drive_conf);
	pci_read_config_byte(dev, (drive_pci), &AP);
	pci_read_config_byte(dev, (drive_pci)|0x01, &BP);
	pci_read_config_byte(dev, (drive_pci)|0x02, &CP);
	pci_read_config_byte(dev, (drive_pci)|0x03, &DP);

	if (speed < XFER_SW_DMA_0) {
		if ((AP & 0x0F) || (BP & 0x07)) {
			/* clear PIO modes of lower 8421 bits of A Register */
			pci_write_config_byte(dev, (drive_pci), AP &~0x0F);
			pci_read_config_byte(dev, (drive_pci), &AP);

			/* clear PIO modes of lower 421 bits of B Register */
			pci_write_config_byte(dev, (drive_pci)|0x01, BP &~0x07);
			pci_read_config_byte(dev, (drive_pci)|0x01, &BP);

			pci_read_config_byte(dev, (drive_pci), &AP);
			pci_read_config_byte(dev, (drive_pci)|0x01, &BP);
		}
#ifdef CONFIG_BLK_DEV_IDEDMA
	} else {
		if ((BP & 0xF0) && (CP & 0x0F)) {
			/* clear DMA modes of upper 842 bits of B Register */
			/* clear PIO forced mode upper 1 bit of B Register */
			pci_write_config_byte(dev, (drive_pci)|0x01, BP &~0xF0);
			pci_read_config_byte(dev, (drive_pci)|0x01, &BP);

			/* clear DMA modes of lower 8421 bits of C Register */
			pci_write_config_byte(dev, (drive_pci)|0x02, CP &~0x0F);
			pci_read_config_byte(dev, (drive_pci)|0x02, &CP);
		}
#endif /* CONFIG_BLK_DEV_IDEDMA */
	}

	pci_read_config_byte(dev, (drive_pci), &AP);
	pci_read_config_byte(dev, (drive_pci)|0x01, &BP);
	pci_read_config_byte(dev, (drive_pci)|0x02, &CP);

	switch(speed) {
#ifdef CONFIG_BLK_DEV_IDEDMA
		/* case XFER_UDMA_6: */
		case XFER_UDMA_5:
		case XFER_UDMA_4:	TB = 0x20; TC = 0x01; break;
		case XFER_UDMA_3:	TB = 0x40; TC = 0x02; break;
		case XFER_UDMA_2:	TB = 0x20; TC = 0x01; break;
		case XFER_UDMA_1:	TB = 0x40; TC = 0x02; break;
		case XFER_UDMA_0:	TB = 0x60; TC = 0x03; break;
		case XFER_MW_DMA_2:	TB = 0x60; TC = 0x03; break;
		case XFER_MW_DMA_1:	TB = 0x60; TC = 0x04; break;
		case XFER_MW_DMA_0:	TB = 0x60; TC = 0x05; break;
		case XFER_SW_DMA_2:	TB = 0x60; TC = 0x05; break;
		case XFER_SW_DMA_1:	TB = 0x80; TC = 0x06; break;
		case XFER_SW_DMA_0:	TB = 0xC0; TC = 0x0B; break;
#endif /* CONFIG_BLK_DEV_IDEDMA */
		case XFER_PIO_4:	TA = 0x01; TB = 0x04; break;
		case XFER_PIO_3:	TA = 0x02; TB = 0x06; break;
		case XFER_PIO_2:	TA = 0x03; TB = 0x08; break;
		case XFER_PIO_1:	TA = 0x05; TB = 0x0C; break;
		case XFER_PIO_0:
		default:		TA = 0x09; TB = 0x13; break;
	}

	if (speed < XFER_SW_DMA_0) {
		pci_write_config_byte(dev, (drive_pci), AP|TA);
		pci_write_config_byte(dev, (drive_pci)|0x01, BP|TB);
#ifdef CONFIG_BLK_DEV_IDEDMA
	} else {
		pci_write_config_byte(dev, (drive_pci)|0x01, BP|TB);
		pci_write_config_byte(dev, (drive_pci)|0x02, CP|TC);
#endif /* CONFIG_BLK_DEV_IDEDMA */
	}

#if PDC202XX_DECODE_REGISTER_INFO
	pci_read_config_byte(dev, (drive_pci), &AP);
	pci_read_config_byte(dev, (drive_pci)|0x01, &BP);
	pci_read_config_byte(dev, (drive_pci)|0x02, &CP);
	pci_read_config_byte(dev, (drive_pci)|0x03, &DP);

	decode_registers(REG_A, AP);
	decode_registers(REG_B, BP);
	decode_registers(REG_C, CP);
	decode_registers(REG_D, DP);
#endif /* PDC202XX_DECODE_REGISTER_INFO */
	if (!drive->init_speed)
		drive->init_speed = speed;
#if PDC202XX_DEBUG_DRIVE_INFO
	printk("%s: Before set_feature = %s, Word88 = %#x\n",
		drive->name, ide_xfer_verbose(speed), drive->id->dma_ultra );
#endif /* PDC202XX_DEBUG_DRIVE_INFO */
	if (drive->present)
		err = ide_config_drive_speed(drive, speed);
	else
		err = 0;
	drive->current_speed = speed;
	
	for ( i = 0 ; i < 2 ; i++)	/* get channel identify info */
		if (hwif->drives[i].present)
	  		id[ i+j+k ] = hwif->drives[i].id;
	/* i = master/slave, j = channel, k = adapter */
	
#if PDC202XX_DEBUG_DRIVE_INFO
	printk("%s: %s drive%d 0x%08x ",
		drive->name, ide_xfer_verbose(speed),
		drive->dn, drive_conf);
		pci_read_config_dword(dev, drive_pci, &drive_conf);
	printk("0x%08x\n", drive_conf);
#endif /* PDC202XX_DEBUG_DRIVE_INFO */
	return err;
}

#define set_2regs(a, b) \
do { OUT_BYTE((a + adj), indexreg); OUT_BYTE(b, datareg); } while (0) ;

#define set_reg_and_wait(value, reg, delay) \
do { OUT_BYTE(value, reg); mdelay(delay); } while (0) ;

static int pdc202xx_new_tune_chipset (ide_drive_t *drive, byte xferspeed)
{
	ide_hwif_t *hwif	= HWIF(drive);
	struct pci_dev *dev	= hwif->pci_dev;
	u32 high_16		= pci_resource_start(dev, 4);
	u32 indexreg		= hwif->channel ? high_16 + 0x09 : high_16 + 0x01;
	u32 datareg		= indexreg + 2;
	byte thold		= 0x10;
	byte adj		= (drive->dn%2) ? 0x08 : 0x00;
	byte speed		= pdc202xx_ratefilter(drive, xferspeed);
	static int speed_rate[4];
	unsigned int 		err, i = 0, j = hwif->channel ? 2 : 0;
	unsigned int k		= ((hwif->index > 3) & (hwif->index < 6)) ? 4 : 0 ;

#ifdef CONFIG_BLK_DEV_IDEDMA
	/* Setting tHOLD bit to 0 if using UDMA mode 2 */
	if (speed == XFER_UDMA_2)
		set_2regs(thold, (IN_BYTE(datareg) & 0x7f));
#endif /* CONFIG_BLK_DEV_IDEDMA */

#if PDC202XX_DEBUG_DRIVE_INFO
	printk("%s: Before set_feature = %s, Word88 = %#x\n",
		drive->name, ide_xfer_verbose(speed), drive->id->dma_ultra );
#endif /* PDC202XX_DEBUG_DRIVE_INFO */
	if (drive->present)
		err = ide_config_drive_speed(drive, speed);
	else
		err = 0;
	
	drive->current_speed = speed;
	if (!drive->init_speed)
		drive->init_speed = speed;
	
	for ( i = 0 ; i < 2 ; i++)	/* get channel identify info */
		if (hwif->drives[i].present) {
	  		id[ i+j+k ] = hwif->drives[i].id;
			/* i = master/slave, j = channel, k = adapter */
	 		speed_rate[ i+j ] = drive->current_speed;
	 	}
	
	/* We need to set ATA133 timing if ATA133 drives exist */
	if (speed < XFER_UDMA_6)
		return err;
	
	for (i=0; i<4; i++) {
		if (id[i] == NULL)
			continue;
		
		/* Set indexreg to current channel */
		indexreg = (i>1) ? high_16 + 0x09 : high_16 + 0x01;
		datareg	= indexreg + 2;
	
		/* Set adj to current drive */
		adj = (i%2) ? 0x08 : 0x00;
		switch(speed_rate[i]) {
#ifdef CONFIG_BLK_DEV_IDEDMA
			case XFER_UDMA_6:
				set_2regs(0x10, 0x1a);
				set_2regs(0x11, 0x01);
				set_2regs(0x12, 0xcb);
				break;
			case XFER_UDMA_5:
				set_2regs(0x10, 0x1a);
				set_2regs(0x11, 0x02);
				set_2regs(0x12, 0xcb);
				break;
			case XFER_UDMA_4:
				set_2regs(0x10, 0x1a);
				set_2regs(0x11, 0x03);
				set_2regs(0x12, 0xcd);
				break;
			case XFER_UDMA_3:
				set_2regs(0x10, 0x1a);
				set_2regs(0x11, 0x05);
				set_2regs(0x12, 0xcd);
				break;	
			case XFER_UDMA_2:
				set_2regs(0x10, 0x2a);
				set_2regs(0x11, 0x07);
				set_2regs(0x12, 0xcd);
				break;
			case XFER_UDMA_1:
				set_2regs(0x10, 0x3a);
				set_2regs(0x11, 0x0a);
				set_2regs(0x12, 0xd0);
				break;
			case XFER_UDMA_0:
				set_2regs(0x10, 0x4a);
				set_2regs(0x11, 0x0f);
				set_2regs(0x12, 0xd5);
				break;
			case XFER_MW_DMA_2:
				set_2regs(0x0e, 0x69);
				set_2regs(0x0f, 0x25);
				break;
			case XFER_MW_DMA_1:
				set_2regs(0x0e, 0x6b);
				set_2regs(0x0f, 0x27);
				break;
			case XFER_MW_DMA_0:
				set_2regs(0x0e, 0xdf);
				set_2regs(0x0f, 0x5f);
				break;
#endif /* CONFIG_BLK_DEV_IDEDMA */
			case XFER_PIO_4:
				set_2regs(0x0c, 0x23);
				set_2regs(0x0d, 0x09);
				set_2regs(0x13, 0x25);
				break;
			case XFER_PIO_3:
				set_2regs(0x0c, 0x27);
				set_2regs(0x0d, 0x0d);
				set_2regs(0x13, 0x35);
				break;
			case XFER_PIO_2:
				set_2regs(0x0c, 0x23);
				set_2regs(0x0d, 0x26);
				set_2regs(0x13, 0x64);
				break;
			case XFER_PIO_1:
				set_2regs(0x0c, 0x46);
				set_2regs(0x0d, 0x29);
				set_2regs(0x13, 0xa4);
				break;
			case XFER_PIO_0:
				set_2regs(0x0c, 0xfb);
				set_2regs(0x0d, 0x2b);
				set_2regs(0x13, 0xac);
				break;
			default:
				;
		}
	}
	return err;
}

/*   0    1    2    3    4    5    6   7   8
 * 960, 480, 390, 300, 240, 180, 120, 90, 60
 *           180, 150, 120,  90,  60
 * DMA_Speed
 * 180, 120,  90,  90,  90,  60,  30
 *  11,   5,   4,   3,   2,   1,   0
 */
static int config_chipset_for_pio (ide_drive_t *drive, byte pio)
{
	byte speed = 0x00;

	pio = (pio == 5) ? 4 : pio;
	speed = XFER_PIO_0 + ide_get_best_pio_mode(drive, 255, pio, NULL);
        
	return ((int) pdc202xx_tune_chipset(drive, speed));
}

static void pdc202xx_tune_drive (ide_drive_t *drive, byte pio)
{
	(void) config_chipset_for_pio(drive, pio);
}

#ifdef CONFIG_BLK_DEV_IDEDMA
static int config_chipset_for_dma (ide_drive_t *drive)
{
	struct hd_driveid *id	= drive->id;
//	byte mode               = pdc202xx_ratemask(drive);
	ide_hwif_t *hwif	= HWIF(drive);
	struct pci_dev *dev	= hwif->pci_dev;
	u32 high_16		= pci_resource_start(dev, 4);
	u32 dma_base		= hwif->dma_base;
	u32 indexreg		= dma_base + 1;
	u32 datareg		= dma_base + 3;
	byte iordy		= 0x13;
	byte adj		= (drive->dn%2) ? 0x08 : 0x00;
	unsigned int cable	= 0;
	byte tx2_family		= 0;
	byte unit		= (drive->select.b.unit & 0x01);
	unsigned int drive_conf;
	byte drive_pci 		= 0;
	byte			test1, test2, speed = -1;
	byte			AP;
	unsigned short		EP;
	u32 clockreg		= high_16 + 0x11;
	byte CLKSPD		= IN_BYTE(clockreg);
	byte udma_33		= 1;
	byte udma_66		= (eighty_ninty_three(drive)) ? 1 : 0;
	byte udma_100		= 0;
	byte udma_133		= 0;
	byte mask		= hwif->channel ? 0x08 : 0x02;
	unsigned short c_mask	= hwif->channel ? (1<<11) : (1<<10);

	byte ultra_66		= ((id->dma_ultra & 0x0010) ||
				   (id->dma_ultra & 0x0008)) ? 1 : 0;
	byte ultra_100		= ((id->dma_ultra & 0x0020) ||
				   (ultra_66)) ? 1 : 0;
	byte ultra_133		= ((id->dma_ultra & 0x0040) ||
				   (ultra_100)) ? 1 : 0;

	switch(dev->device) {
		case PCI_DEVICE_ID_PROMISE_20277:
		case PCI_DEVICE_ID_PROMISE_20276:
		case PCI_DEVICE_ID_PROMISE_20275:
		case PCI_DEVICE_ID_PROMISE_20271:
		case PCI_DEVICE_ID_PROMISE_20269:
			udma_100 = (udma_66) ? 1 : 0;
			udma_133 = (udma_100) ? 1 : 0;
		case PCI_DEVICE_ID_PROMISE_20270:
		case PCI_DEVICE_ID_PROMISE_20268:
			udma_100 = (udma_66) ? 1 : 0;
			OUT_BYTE(0x0b, indexreg);
			cable = (IN_BYTE(datareg) & 0x04);
			/* 0 = 80-pin cable */
			tx2_family = 1;
			break;
		case PCI_DEVICE_ID_PROMISE_20267:
		case PCI_DEVICE_ID_PROMISE_20265:
			udma_100 = (udma_66) ? 1 : 0;
		case PCI_DEVICE_ID_PROMISE_20263:
		case PCI_DEVICE_ID_PROMISE_20262:
			pci_read_config_word(dev, 0x50, &EP);
			cable = (EP & c_mask);
			/* 0 = 80-pin cable */
			tx2_family = 0;
			break;
		default:
			udma_100 = 0; udma_133 = 0; 
			cable = 1; tx2_family = 0;
			break;
	}

	/*
	 * Set the control register to use the 66Mhz system
	 * clock for UDMA 3/4 mode operation. If one drive on
	 * a channel is U66 capable but the other isn't we
	 * fall back to U33 mode. The BIOS INT 13 hooks turn
	 * the clock on then off for each read/write issued. I don't
	 * do that here because it would require modifying the
	 * kernel, seperating the fop routines from the kernel or
	 * somehow hooking the fops calls. It may also be possible to
	 * leave the 66Mhz clock on and readjust the timing
	 * parameters.
	 */
	if ( ultra_66 && cable ) {
#ifdef DEBUG
		printk("ULTRA 66/100/133: %s channel of Ultra 66/100/133 "
			"requires an 80-pin cable for Ultra66 operation.\n",
			hwif->channel ? "Secondary" : "Primary");
		printk("         Switching to Ultra33 mode.\n");
#endif /* DEBUG */
		if (!tx2_family)
			OUT_BYTE(CLKSPD & ~mask, clockreg);
		printk("pdc202xx: %s channel use 40-pin cable, reduced to ATA-33 mode!!\n", hwif->channel ? "Secondary" : "Primary");
		udma_66 = 0; udma_100 = 0; udma_133 = 0;
	} else {
		if ((ultra_66) || (ultra_100) || (ultra_133)) {
			/*
			 * check to make sure drive on same channel
			 * is u66 capable
			 */
			if (hwif->drives[!(drive->dn%2)].id) {
				if (hwif->drives[!(drive->dn%2)].id->dma_ultra & 0x0078) {
					if (!tx2_family)
						OUT_BYTE(CLKSPD | mask, clockreg);
				}
				else {
					if (!tx2_family)
						OUT_BYTE(CLKSPD & ~mask, clockreg);
				}
			} else {/* udma4 drive by itself */
				if (!tx2_family)
					OUT_BYTE(CLKSPD | mask, clockreg);
			}
		}
	}

	if (tx2_family)	goto chipset_is_set;

	switch(drive->dn) {
		case 0:	drive_pci = 0x60;
			pci_read_config_dword(dev, drive_pci, &drive_conf);
			if ((drive_conf != 0x004ff304) && (drive_conf != 0x004ff3c4))
				goto chipset_is_set;
			pci_read_config_byte(dev, (drive_pci), &test1);
			if (!(test1 & SYNC_ERRDY_EN))
				pci_write_config_byte(dev, (drive_pci), test1|SYNC_ERRDY_EN);
			break;
		case 1:	drive_pci = 0x64;
			pci_read_config_dword(dev, drive_pci, &drive_conf);
			if ((drive_conf != 0x004ff304) && (drive_conf != 0x004ff3c4))
				goto chipset_is_set;
			pci_read_config_byte(dev, 0x60, &test1);
			pci_read_config_byte(dev, (drive_pci), &test2);
			if ((test1 & SYNC_ERRDY_EN) && !(test2 & SYNC_ERRDY_EN))
				pci_write_config_byte(dev, (drive_pci), test2|SYNC_ERRDY_EN);
			break;
		case 2:	drive_pci = 0x68;
			pci_read_config_dword(dev, drive_pci, &drive_conf);
			if ((drive_conf != 0x004ff304) && (drive_conf != 0x004ff3c4))
				goto chipset_is_set;
			pci_read_config_byte(dev, (drive_pci), &test1);
			if (!(test1 & SYNC_ERRDY_EN))
				pci_write_config_byte(dev, (drive_pci), test1|SYNC_ERRDY_EN);
			break;
		case 3:	drive_pci = 0x6c;
			pci_read_config_dword(dev, drive_pci, &drive_conf);
			if ((drive_conf != 0x004ff304) && (drive_conf != 0x004ff3c4))
				goto chipset_is_set;
			pci_read_config_byte(dev, 0x68, &test1);
			pci_read_config_byte(dev, (drive_pci), &test2);
			if ((test1 & SYNC_ERRDY_EN) && !(test2 & SYNC_ERRDY_EN))
				pci_write_config_byte(dev, (drive_pci), test2|SYNC_ERRDY_EN);
			break;
		default:
			return ide_dma_off;
	}

chipset_is_set:

	if (drive->media != ide_disk) {
		hwif->tuneproc(drive, 5);
		return ide_dma_off_quietly;
	}

	if (tx2_family) {
		if (id->capability & 4)	/* IORDY_EN & PREFETCH_EN */
			set_2regs(iordy, (IN_BYTE(datareg) | 0x03));
	}
	else {
		pci_read_config_byte(dev, (drive_pci), &AP);
		if (id->capability & 4)		/* IORDY_EN */
			pci_write_config_byte(dev, (drive_pci), AP|IORDY_EN);
		pci_read_config_byte(dev, (drive_pci), &AP);
		if (drive->media == ide_disk)	/* PREFETCH_EN */
			pci_write_config_byte(dev, (drive_pci), AP|PREFETCH_EN);
	}

	if ((id->dma_ultra & 0x0040)&&(udma_133))	speed = XFER_UDMA_6;
	else if ((id->dma_ultra & 0x0020)&&(udma_100))	speed = XFER_UDMA_5;
	else if ((id->dma_ultra & 0x0010)&&(udma_66))	speed = XFER_UDMA_4;
	else if ((id->dma_ultra & 0x0008)&&(udma_66))	speed = XFER_UDMA_3;
	else if ((id->dma_ultra & 0x0004)&&(udma_33))	speed = XFER_UDMA_2;
	else if ((id->dma_ultra & 0x0002)&&(udma_33))	speed = XFER_UDMA_1;
	else if ((id->dma_ultra & 0x0001)&&(udma_33))	speed = XFER_UDMA_0;
	else if (id->dma_mword & 0x0004)		speed = XFER_MW_DMA_2;
	else if (id->dma_mword & 0x0002)		speed = XFER_MW_DMA_1;
	else if (id->dma_mword & 0x0001)		speed = XFER_MW_DMA_0;
	else if ((id->dma_1word & 0x0004)&&(!tx2_family))
		speed = XFER_SW_DMA_2;
	else if ((id->dma_1word & 0x0002)&&(!tx2_family))
		speed = XFER_SW_DMA_1;
	else if ((id->dma_1word & 0x0001)&&(!tx2_family))
		speed = XFER_SW_DMA_0;
	else {
		/* restore original pci-config space */
		if (!tx2_family)
			pci_write_config_dword(dev, drive_pci, drive_conf);
		hwif->tuneproc(drive, 5);
		return ide_dma_off_quietly;
	}

	OUT_BYTE(IN_BYTE(dma_base+2) & ~(1<<(5+unit)), dma_base+2);
	(void) hwif->speedproc(drive, speed);

	return ((int)	((id->dma_ultra >> 14) & 3) ? ide_dma_on :
			((id->dma_ultra >> 11) & 7) ? ide_dma_on :
			((id->dma_ultra >> 8) & 7) ? ide_dma_on :
			((id->dma_mword >> 8) & 7) ? ide_dma_on : 
			((id->dma_1word >> 8) & 7) ? ide_dma_on :
						     ide_dma_off_quietly);
}

static int config_drive_xfer_rate (ide_drive_t *drive)
{
	struct hd_driveid *id		= drive->id;
	ide_hwif_t *hwif		= HWIF(drive);
	ide_dma_action_t dma_func	= ide_dma_off_quietly;

	drive->init_speed = 0;

	if (id && (id->capability & 1) && hwif->autodma) {
		/* Consult the list of known "bad" drives */
		if (ide_dmaproc(ide_dma_bad_drive, drive)) {
			dma_func = ide_dma_off;
			goto fast_ata_pio;
		}
		dma_func = ide_dma_off_quietly;
		if (id->field_valid & 4) {
			if (id->dma_ultra & 0x007F) {
				/* Force if Capable UltraDMA */
				dma_func = config_chipset_for_dma(drive);
				if ((id->field_valid & 2) &&
				    (dma_func != ide_dma_on))
					goto try_dma_modes;
			}
		} else if (id->field_valid & 2) {
try_dma_modes:
			if ((id->dma_mword & 0x0007) ||
			    (id->dma_1word & 0x0007)) {
				/* Force if Capable regular DMA modes */
				dma_func = config_chipset_for_dma(drive);
				if (dma_func != ide_dma_on)
					goto no_dma_set;
			}
		} else if (ide_dmaproc(ide_dma_good_drive, drive)) {
			if (id->eide_dma_time > 150) {
				goto no_dma_set;
			}
			/* Consult the list of known "good" drives */
			dma_func = config_chipset_for_dma(drive);
			if (dma_func != ide_dma_on)
				goto no_dma_set;
		} else {
			goto fast_ata_pio;
		}
	} else if ((id->capability & 8) || (id->field_valid & 2)) {
fast_ata_pio:
		dma_func = ide_dma_off_quietly;
no_dma_set:
		hwif->tuneproc(drive, 5);
	}

	return hwif->dmaproc(dma_func, drive);
}

int pdc202xx_quirkproc (ide_drive_t *drive)
{
	return ((int) check_in_drive_lists(drive, pdc_quirk_drives));
}

/*
 * pdc202xx_dmaproc() initiates/aborts (U)DMA read/write operations on a drive.
 */
int pdc202xx_dmaproc (ide_dma_action_t func, ide_drive_t *drive)
{
	byte dma_stat		= 0;
	byte sc1d		= 0;
	byte tx2_family		= 0;
	byte clock		= 0;
	byte hardware48fix	= 0;
	ide_hwif_t *hwif	= HWIF(drive);
	struct pci_dev *dev	= hwif->pci_dev;
	u32 high_16		= pci_resource_start(dev, 4);
	u32 clockreg		= high_16 + 0x11;
	u32 atapi_port		= hwif->channel ? high_16 + 0x24 : high_16 + 0x20;
	u32 dma_base		= hwif->dma_base;

	switch (dev->device) {
		case PCI_DEVICE_ID_PROMISE_20277:
		case PCI_DEVICE_ID_PROMISE_20276:
		case PCI_DEVICE_ID_PROMISE_20275:
		case PCI_DEVICE_ID_PROMISE_20271:
		case PCI_DEVICE_ID_PROMISE_20270:
		case PCI_DEVICE_ID_PROMISE_20269:
		case PCI_DEVICE_ID_PROMISE_20268:
			tx2_family = 1;
			break;
		case PCI_DEVICE_ID_PROMISE_20267:
		case PCI_DEVICE_ID_PROMISE_20265:
		case PCI_DEVICE_ID_PROMISE_20262:
		/* PDC20262 acturally do not support 48-bit LBA */
			hardware48fix = 1;
			clock = IN_BYTE(clockreg);
		default:
			break;
	}

	switch (func) {
		case ide_dma_check:
			return config_drive_xfer_rate(drive);
		case ide_dma_begin:
			/* Note that this is done *after* the cmd has
			 * been issued to the drive, as per the BM-IDE spec.
			 * The Promise Ultra33 doesn't work correctly when
			 * we do this part before issuing the drive cmd.
			 */
			/* Enable ATAPI UDMA port for LBA48 on PDC20265/267 */
			if ((drive->addressing == 1) && (hardware48fix)) {
				struct request *rq = HWGROUP(drive)->rq;
				u32 word_count = 0;
				u32 hankval = 0;

				OUT_BYTE(clock|(hwif->channel ? 0x08:0x02), clockreg);
				word_count = (rq->nr_sectors << 8);
				hankval = (rq->cmd == READ) ? 0x05<<24 : 0x06<<24;
				hankval = hankval | word_count ;
				outl(hankval, atapi_port);
			}
			break;
		case ide_dma_end:
			/* Disable ATAPI UDMA port for LBA48 on PDC20265/267 */
			if ((drive->addressing == 1) && (hardware48fix)) {
				outl(0, atapi_port);	/* zero out extra */
				clock = IN_BYTE(clockreg);
				OUT_BYTE(clock & ~(hwif->channel ? 0x08:0x02), clockreg);
			}
			break;
		case ide_dma_test_irq:	/* returns 1 if dma irq issued, 0 otherwise */
			dma_stat = IN_BYTE(dma_base+2);
			if (tx2_family)
				return (dma_stat & 4) == 4;

			sc1d = IN_BYTE(high_16 + 0x001d);
			if (hwif->channel) {
				if ((sc1d & 0x50) == 0x50) goto somebody_else;
				else if ((sc1d & 0x40) == 0x40)
					return (dma_stat & 4) == 4;
			} else {
				if ((sc1d & 0x05) == 0x05) goto somebody_else;
				else if ((sc1d & 0x04) == 0x04)
					return (dma_stat & 4) == 4;
			}
somebody_else:
			return (dma_stat & 4) == 4;	/* return 1 if INTR asserted */
		case ide_dma_lostirq:
		case ide_dma_timeout:
			if (hwif->resetproc != NULL)
				hwif->resetproc(drive);
		default:
			break;
	}
	return ide_dmaproc(func, drive);	/* use standard DMA stuff */
}
#endif /* CONFIG_BLK_DEV_IDEDMA */

void pdc202xx_reset (ide_drive_t *drive)
{
	u32 high_16		= pci_resource_start(HWIF(drive)->pci_dev, 4);
	byte udma_speed_flag	= IN_BYTE(high_16 + 0x001f);

	set_reg_and_wait(udma_speed_flag | 0x10, high_16+0x001f, 100);
	/* FIXME: 2 seconds ?! */
	set_reg_and_wait(udma_speed_flag & ~0x10, high_16+0x001f, 2000);
	
	printk("pdc202xx: drive %s reset done.\n",drive->name);
}

void pdc202xx_new_reset (ide_drive_t *drive)
{
	u32 high_16	= pci_resource_start(HWIF(drive)->pci_dev, 4);
	u32 indexreg	= HWIF(drive)->channel ? high_16+0x09 : high_16+0x01;
	u32 datareg	= indexreg + 2;
	byte hrst_bit	= HWIF(drive)->channel ? 0xFB : 0xF7, status;
	
	OUT_BYTE( 1, indexreg );
	status = IN_BYTE( datareg );
	OUT_BYTE(  (status & hrst_bit) , datareg );
	udelay(10);
	OUT_BYTE( status , datareg );
	mdelay(200);
	
	printk("pdc202xx: drive %s reset done.\n",drive->name);
}

void pdc202xx_marvell_idle (ide_drive_t *drive)
{
	/* Give a breath for Marvell's Idle Immediate command */
	OUT_BYTE(0, IDE_NSECTOR_REG); /* timeout disable */
	OUT_BYTE(WIN_SETIDLE1, IDE_COMMAND_REG);
}

/*
 * Since SUN Cobalt is attempting to do this operation, I should disclose
 * this has been a long time ago Thu Jul 27 16:40:57 2000 was the patch date
 * HOTSWAP ATA Infrastructure.
 */
static int pdc202xx_tristate (ide_drive_t * drive, int state)
{
#if 0
	ide_hwif_t *hwif	= HWIF(drive);
	unsigned long high_16	= pci_resource_start(hwif->pci_dev, 4);
	byte sc1f		= IN_BYTE(high_16 + 0x001f);

	if (!hwif)
		return -EINVAL;

//	hwif->bus_state = state;

	if (state) {
		OUT_BYTE(sc1f | 0x08, high_16 + 0x001f);
	} else {
		OUT_BYTE(sc1f & ~0x08, high_16 + 0x001f);
	}
#endif
	return 0;
}

unsigned int __init pci_init_pdc202xx (struct pci_dev *dev, const char *name)
{
	u32 high_16		= pci_resource_start(dev, 4);
	byte tx2_family		= 0;
	byte udma_speed_flag	= 0;
	byte primary_mode	= 0;
	byte secondary_mode	= 0;

	switch (dev->device) {
		case PCI_DEVICE_ID_PROMISE_20277:
		case PCI_DEVICE_ID_PROMISE_20276:
		case PCI_DEVICE_ID_PROMISE_20275:
		case PCI_DEVICE_ID_PROMISE_20271:
		case PCI_DEVICE_ID_PROMISE_20269:
		case PCI_DEVICE_ID_PROMISE_20270:
		case PCI_DEVICE_ID_PROMISE_20268:
			tx2_family = 1;
			break;
		case PCI_DEVICE_ID_PROMISE_20267:
		case PCI_DEVICE_ID_PROMISE_20265:
		case PCI_DEVICE_ID_PROMISE_20263:
		case PCI_DEVICE_ID_PROMISE_20262:
			udma_speed_flag	= IN_BYTE(high_16 + 0x001f);
			primary_mode	= IN_BYTE(high_16 + 0x001a);
			secondary_mode	= IN_BYTE(high_16 + 0x001b);
			/*
			 * software reset -  this is required because the bios
			 * will set UDMA timing on if the hdd supports it. The
			 * user may want to turn udma off. A bug in the pdc20262
			 * is that it cannot handle a downgrade in timing from
			 * UDMA to DMA. Disk accesses after issuing a set
			 * feature command will result in errors. A software
			 * reset leaves the timing registers intact,
			 * but resets the drives.
			 */
			set_reg_and_wait(udma_speed_flag | 0x10, high_16 + 0x001f, 100);
			/* FIXME: 2 seconds ?! */
			set_reg_and_wait(udma_speed_flag & ~0x10, high_16 + 0x001f, 2000);
		default:
			break;
	}
	
	if (dev->resource[PCI_ROM_RESOURCE].start) {
		pci_write_config_dword(dev, PCI_ROM_ADDRESS,
			dev->resource[PCI_ROM_RESOURCE].start | PCI_ROM_ADDRESS_ENABLE);
		printk("%s: ROM enabled at 0x%08lx\n",
			name, dev->resource[PCI_ROM_RESOURCE].start);
	}

	if (tx2_family) goto skip_burst;

	printk("%s: (U)DMA Burst Bit %sABLED " \
		"Primary %s Mode " \
		"Secondary %s Mode.\n",
		name,
		(udma_speed_flag & 1) ? "EN" : "DIS",
		(primary_mode & 1) ? "MASTER" : "PCI",
		(secondary_mode & 1) ? "MASTER" : "PCI" );

#ifdef CONFIG_PDC202XX_BURST
	if (!(udma_speed_flag & 1)) {
		printk("%s: FORCING BURST BIT 0x%02x -> 0x%02x ", name, udma_speed_flag, (udma_speed_flag|1));
		OUT_BYTE(udma_speed_flag|1, high_16 + 0x001f);
		printk("%sCTIVE\n", (IN_BYTE(high_16 + 0x001f) & 1) ? "A" : "INA");
	}
#endif /* CONFIG_PDC202XX_BURST */

#ifdef CONFIG_PDC202XX_MASTER
	if (!(primary_mode & 1)) {
		printk("%s: FORCING PRIMARY MODE BIT 0x%02x -> 0x%02x ",
			name, primary_mode, (primary_mode|1));
		OUT_BYTE(primary_mode|1, high_16 + 0x001a);
		printk("%s\n", (IN_BYTE(high_16 + 0x001a) & 1) ? "MASTER" : "PCI");
	}

	if (!(secondary_mode & 1)) {
		printk("%s: FORCING SECONDARY MODE BIT 0x%02x -> 0x%02x ",
			name, secondary_mode, (secondary_mode|1));
		OUT_BYTE(secondary_mode|1, high_16 + 0x001b);
		printk("%s\n", (IN_BYTE(high_16 + 0x001b) & 1) ? "MASTER" : "PCI");
	}
#endif /* CONFIG_PDC202XX_MASTER */
	
skip_burst:
	pdc202_devs[n_pdc202_devs++] = dev;

#if defined(DISPLAY_PDC202XX_TIMINGS) && defined(CONFIG_PROC_FS)
	if (!pdc202xx_proc) {
		pdc202xx_proc = 1;
		pdc202xx_display_info = &pdc202xx_get_info;
	}
#endif /* DISPLAY_PDC202XX_TIMINGS && CONFIG_PROC_FS */

	return dev->irq;
}

unsigned int __init ata66_pdc202xx (ide_hwif_t *hwif)
{
	unsigned short mask = (hwif->channel) ? (1<<11) : (1<<10);
	unsigned short CIS;

        switch(hwif->pci_dev->device) {
		case PCI_DEVICE_ID_PROMISE_20277:
		case PCI_DEVICE_ID_PROMISE_20276:
		case PCI_DEVICE_ID_PROMISE_20275:
		case PCI_DEVICE_ID_PROMISE_20271:
		case PCI_DEVICE_ID_PROMISE_20269:
		case PCI_DEVICE_ID_PROMISE_20268:
		case PCI_DEVICE_ID_PROMISE_20270:
			/* check 80pin cable */
			OUT_BYTE(0x0b, ((hwif->dma_base) + 1));
			return (!(IN_BYTE(((hwif->dma_base) + 3)) & 0x04));
		default:
			/* check 80pin cable */
			pci_read_config_word(hwif->pci_dev, 0x50, &CIS);
			return (!(CIS & mask));
	}
	/* hwif->udma_four = ata66_pdc202xx() ;
	 * So we need to inverse the value */
}

void __init ide_init_pdc202xx (ide_hwif_t *hwif)
{
	int i		= 0;
	hwif->tuneproc  = &pdc202xx_tune_drive;
	hwif->quirkproc = &pdc202xx_quirkproc;

	for (i=0; i<2; i++)
		hwif->drives[i].autotune = 1;

	hwif->autodma = 0;

#undef CONFIG_PDC202XX_32_UNMASK
#ifdef CONFIG_PDC202XX_32_UNMASK
	for (i=0; i<2; i++) {
		hwif->drives[i].io_32bit = 1;
		hwif->drives[i].unmask = 1;
	}
#endif /* CONFIG_PDC202XX_32_UNMASK */

        switch(hwif->pci_dev->device) {
		case PCI_DEVICE_ID_PROMISE_20277:
		case PCI_DEVICE_ID_PROMISE_20276:
		case PCI_DEVICE_ID_PROMISE_20275:
		case PCI_DEVICE_ID_PROMISE_20271:
		case PCI_DEVICE_ID_PROMISE_20269:
		case PCI_DEVICE_ID_PROMISE_20268:
		case PCI_DEVICE_ID_PROMISE_20270:
			hwif->resetproc	= &pdc202xx_new_reset;
			hwif->speedproc = &pdc202xx_new_tune_chipset;
			break;
		case PCI_DEVICE_ID_PROMISE_20267:
		case PCI_DEVICE_ID_PROMISE_20265:
		case PCI_DEVICE_ID_PROMISE_20263:
		case PCI_DEVICE_ID_PROMISE_20262:
			hwif->resetproc	= &pdc202xx_reset;
			hwif->busproc   = &pdc202xx_tristate;
		case PCI_DEVICE_ID_PROMISE_20246:
			hwif->speedproc = &pdc202xx_tune_chipset;
		default:
			break;
	}

	if (!hwif->dma_base)
		return;

#ifdef CONFIG_BLK_DEV_IDEDMA
	hwif->dmaproc = &pdc202xx_dmaproc;
# ifdef CONFIG_IDEDMA_AUTO
	if (!noautodma)
		hwif->autodma = 1;
# endif /* CONFIG_IDEDMA_AUTO */
#endif /* CONFIG_BLK_DEV_IDEDMA */
}
/*
extern void ide_setup_pci_device (struct pci_dev *dev, ide_pci_device_t *d);

void __init fixup_device_pdc20265 (struct pci_dev *dev, ide_pci_device_t *d)
{
	printk("%s: IDE controller on PCI bus %02x dev %02x\n",
		d->name, dev->bus->number, dev->devfn);
	ide_setup_pci_device(dev, d);
}

void __init fixup_device_pdc20270 (struct pci_dev *dev, ide_pci_device_t *d)
{
	struct pci_dev *dev2 = NULL, *findev;
	ide_pci_device_t *d2;

	if ((dev->bus->self &&
	     dev->bus->self->vendor == PCI_VENDOR_ID_DEC) &&
	    (dev->bus->self->device == PCI_DEVICE_ID_DEC_21150)) {
		if (PCI_SLOT(dev->devfn) & 2) {
			return;
		}
		d->extra = 0;
		pci_for_each_dev(findev) {
			if ((findev->vendor == dev->vendor) &&
			    (findev->device == dev->device) &&
			    (PCI_SLOT(findev->devfn) & 2)) {
				byte irq = 0, irq2 = 0;
				dev2 = findev;
				pci_read_config_byte(dev, PCI_INTERRUPT_LINE, &irq);
				pci_read_config_byte(dev2, PCI_INTERRUPT_LINE, &irq2);
				if (irq != irq2) {
					dev2->irq = dev->irq;
					pci_write_config_byte(dev2, PCI_INTERRUPT_LINE, irq);
				}
			}
		}
	}

	printk("%s: IDE controller on PCI bus %02x dev %02x\n",
		d->name, dev->bus->number, dev->devfn);
	ide_setup_pci_device(dev, d);
	if (!dev2)
		return;
	d2 = d;
	printk("%s: IDE controller on PCI bus %02x dev %02x\n",
		d2->name, dev2->bus->number, dev2->devfn);
	ide_setup_pci_device(dev2, d2);
}*/
