/*
 *   Copyright (c) International Business Machines  Corp., 2000
 *
 *   This program is free software;  you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY;  without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 *   the GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program;  if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * linux/include/linux/dm-bbr.h
 *
 * Bad-block-relocation (BBR) target for device-mapper.
 *
 * The BBR target is designed to remap I/O write failures to another safe
 * location on disk. Note that most disk drives have BBR built into them,
 * this means that our software BBR will be only activated when all hardware
 * BBR replacement sectors have been used.
 */

#ifndef __DM_BBR__
#define __DM_BBR__

#define BBR_TABLE_SIGNATURE		0x42627254 /* BbrT */
#define BBR_ENTRIES_PER_SECT		31
#define BBR_NR_BUFS			128
#define INITIAL_CRC			0xFFFFFFFF
#define CRC_POLYNOMIAL			0xEDB88320L
#define BBR_STOP_REMAP			(1<<0)

/**
 * Macros to cleanly print 64-bit numbers on both 32-bit and 64-bit machines.
 * Use these in place of %Ld, %Lu, and %Lx.
 **/
#if BITS_PER_LONG > 32
#define PFU64 "%lu"
#else
#define PFU64 "%Lu"
#endif

/**
 * struct bbr_table_entry
 * @bad_sect:		LBA of bad location.
 * @replacement_sect:	LBA of new location.
 *
 * Structure to describe one BBR remap.
 **/
struct bbr_table_entry {
	u64 bad_sect;
	u64 replacement_sect;
};

/**
 * struct bbr_table
 * @signature:		Signature on each BBR table sector.
 * @crc:		CRC for this table sector.
 * @sequence_number:	Used to resolve conflicts when primary and secondary
 *			tables do not match.
 * @in_use_cnt:		Number of in-use table entries.
 * @entries:		Actual table of remaps.
 *
 * Structure to describe each sector of the metadata table. Each sector in this
 * table can describe 31 remapped sectors.
 **/
struct bbr_table {
	u32			signature;
	u32			crc;
	u32			sequence_number;
	u32			in_use_cnt;
	struct bbr_table_entry	entries[BBR_ENTRIES_PER_SECT];
};

/**
 * struct bbr_runtime_remap
 *
 * Node in the binary tree used to keep track of remaps.
 **/
struct bbr_runtime_remap {
	struct bbr_table_entry		remap;
	struct bbr_runtime_remap	*left;
	struct bbr_runtime_remap	*right;
};

/**
 * struct bbr_private
 * @next:			List of all bbr_private structures.
 * @dev:			Info about underlying device.
 * @bbr_table:			Copy of metadata table.
 * @offset:
 * @lba_table1:			LBA of primary BBR table.
 * @lba_table2:			LBA of secondary BBR table.
 * @nr_sects_bbr_table:		Size of each BBR table.
 * @nr_replacement_blks:	Number of replacement sectors.
 * @start_replacement_sect:	LBA of start of replacement sectors.
 * @blksize_in_sects:		Size of each sector.
 * @in_use_replacement_blks:	Current number of remaps.
 * @remap_root:			Binary tree containing all remaps.
 * @bbr_id_lock:		Lock for the binary tree.
 * @flags:			BBR_STOP_REMAP
 **/
struct bbr_private {
	struct list_head		bbr_list;
	struct dm_dev			*dev;
	struct bbr_table		*bbr_table;
	u64				offset;
	u64				lba_table1;
	u64				lba_table2;
	u64				nr_sects_bbr_table;
	u64				nr_replacement_blks;
	u64				start_replacement_sect;
	u32				blksize_in_sects;
	atomic_t			in_use_replacement_blks;
	struct bbr_runtime_remap	*remap_root;
	spinlock_t			bbr_id_lock;
	u32				flag;
};

#define BBR_IO_HANDLED		(1<<0)
#define BBR_IO_RELOCATE		(1<<1)

/**
 * struct bbr_io_buffer
 * @bbr_io_list:	Thread's list of bbr_io_buf's.
 * @bbr_id:		Object for this request.
 * @bh:			Original buffer_head.
 * @sector:		Original sector
 * @flags:		Operation flag
 * @rw:			READ or WRITE.
 * @rc:			Return code from bbr_io_handler.
 *
 * Structure used to track each write request.
 **/
struct bbr_io_buffer {
	struct list_head	bbr_io_list;
	struct bbr_private	*bbr_id;
	struct buffer_head	*bh;
	u64			sector;
	u32			flags;
	s32			rw;
	s32			rc;
};

#define BBR_THREAD_WAKEUP 0

/**
 * struct bbr_thread
 * @run:	
 * @data:	
 * @wqueue:	thread wait queue
 * @flags:	thread attributes
 * @event:	event completion
 * @tsk:	task info
 * @name:	thread name
 *
 * data structure for creating/managing a kernel thread
 **/
struct bbr_thread {
	void			(*run) (void *data);
	void			*data;
	wait_queue_head_t	wqueue;
	unsigned long		flags;
	struct completion	*event;
	struct task_struct	*tsk;
	const u8		*name;
};

#endif
