/* Connection tracking via netlink socket. Allows for user space
 * protocol helpers and general trouble making from userspace.
 *
 * (C) 2001 by Jay Schulist <jschlst@samba.org>
 * (C) 2002 by Harald Welte <laforge@gnumonks.org>
 *
 * Initial connection tracking via netlink development funded and 
 * generally made possible by Network Robots, Inc. (www.networkrobots.com)
 *
 * Further development of this code funded by Astaro AG (http://www.asaro.com)
 *
 * This software may be used and distributed according to the terms
 * of the GNU General Public License, incorporated herein by reference.
 */

#include <linux/config.h>
#include <linux/module.h>
#include <linux/types.h>
#include <linux/socket.h>
#include <linux/kernel.h>
#include <linux/major.h>
#include <linux/sched.h>
#include <linux/timer.h>
#include <linux/string.h>
#include <linux/sockios.h>
#include <linux/net.h>
#include <linux/fcntl.h>
#include <linux/skbuff.h>
#include <asm/uaccess.h>
#include <asm/system.h>
#include <net/sock.h>
#include <linux/init.h>
#include <linux/netlink.h>
#include <linux/spinlock.h>
#include <linux/rtnetlink.h>

#include <linux/netfilter.h>
#include <linux/netfilter_ipv4.h>
#include <linux/netfilter_ipv4/ip_tables.h>
#include <linux/netfilter_ipv4/ip_conntrack.h>
#include <linux/netfilter_ipv4/ip_conntrack_core.h>

#include <linux/nfnetlink.h>
#include <linux/nfnetlink_conntrack.h>

#define ASSERT_READ_LOCK(x) MUST_BE_READ_LOCKED(&ip_conntrack_lock)
#define ASSERT_WRITE_LOCK(x) MUST_BE_WRITE_LOCKED(&ip_conntrack_lock)
#include <linux/netfilter_ipv4/listhelp.h>

MODULE_LICENSE("GPL");

char ctversion[] = "0.11";
int ct_debug_level = 1;
#define ct_debug(level, format, arg...) \
	if(ct_debug_level > level)  \
		printk(__FUNCTION__ ": " format, ## arg)

/* FIXME: this define is just needed for DUMP_TUPLE */
#define DEBUGP(format, args...)	ct_debug(0, format, ## args)

static struct nfnetlink_subsystem *ctnl_subsys;

static int
ctnetlink_fill_info(struct sk_buff *skb, u32 pid, u32 seq,
		    int event, 
		    int nowait, 
		    const struct ip_conntrack *ct, 
		    const enum ip_conntrack_info *ctinfo, 
		    unsigned char proto,
		    const struct net_device *in,
		    const struct net_device *out)
{
	struct nlmsghdr *nlh;
	struct nfgenmsg *nfmsg;
	struct cta_proto cp;
	unsigned long s;
	unsigned char *b;

	b = skb->tail;
	nlh = NLMSG_PUT(skb, pid, seq, (NFNL_SUBSYS_CTNETLINK<<8)|event, 
			sizeof(struct nfgenmsg));
	nfmsg = NLMSG_DATA(nlh);

	nlh->nlmsg_flags 	= (nowait && pid) ? NLM_F_MULTI : 0;
	nfmsg->nfgen_family	= AF_INET;
	NFA_PUT(skb, CTA_ORIG, sizeof(struct ip_conntrack_tuple), 
		&ct->tuplehash[IP_CT_DIR_ORIGINAL].tuple);
	NFA_PUT(skb, CTA_RPLY, sizeof(struct ip_conntrack_tuple),
		&ct->tuplehash[IP_CT_DIR_REPLY].tuple);
	s = ct->status;
	NFA_PUT(skb, CTA_STATUS, sizeof(unsigned long), &s);
	if (in)
		NFA_PUT(skb, CTA_IIF, IFNAMSIZ, in->name);
	if (out)
		NFA_PUT(skb, CTA_OIF, IFNAMSIZ, out->name);
	if (ctinfo)
		NFA_PUT(skb, CTA_INFO, sizeof(unsigned long), ctinfo);

	cp.num_proto = proto;
	memcpy(&cp.proto, &ct->proto, sizeof (cp.proto));
	NFA_PUT(skb, CTA_PROTOINFO, sizeof(cp), &cp);

	if (ct->helper) {
		struct cta_help ch;

		memcpy(&ch.tuple, &ct->helper->tuple, 
			sizeof(struct ip_conntrack_tuple));
		memcpy(&ch.mask, &ct->helper->mask,
			sizeof(struct ip_conntrack_tuple));
		strncpy((char *)&ch.name, ct->helper->name, sizeof(ch.name));
		memcpy(&ch.help, &ct->help, sizeof(ch.help));
		NFA_PUT(skb, CTA_HELPINFO, sizeof(ch), &ch);
	}

#ifdef CONFIG_IP_NF_NAT_NEEDED
	if (ct->nat.info.initialized && ct->nat.info.num_manips) {
		const struct ip_nat_info *nat = &ct->nat.info;
		struct cta_nat cn;

		cn.num_manips = nat->num_manips;
		memcpy(&cn.manips, &nat->manips, (nat->num_manips 
			* sizeof(struct ip_nat_info_manip)));
		NFA_PUT(skb, CTA_NATINFO, sizeof(struct cta_nat), &cn);
	}
#endif /* CONFIG_IP_NF_NAT_NEEDED */

	nlh->nlmsg_len = skb->tail - b;
	return skb->len;

nlmsg_failure:
nfattr_failure:
	skb_trim(skb, b - skb->data);
	return -1;
}

static inline struct sk_buff *
ctnetlink_event_build_msg(const struct ip_conntrack *ct, 
			  const enum ip_conntrack_info ctinfo, 
			  const unsigned char proto,
			  const struct net_device *in,
			  const struct net_device *out)
{
	struct sk_buff *skb;
	int err;

	skb = alloc_skb(NLMSG_GOODSIZE, GFP_ATOMIC);
	if (!skb)
		return NULL;

	err = ctnetlink_fill_info(skb, 0, 0, CTNL_MSG_NEWCONNTRACK, 1, ct,
				  &ctinfo, proto, in, out);
	if (err <= 0)
		goto nlmsg_failure;
	return skb;

nlmsg_failure:
	return NULL;
}

static void
ctnetlink_create(struct ip_conntrack *ct,
		 enum ip_conntrack_info ctinfo,
		 const struct net_device *in, 
		 const struct net_device *out)
{
	u16 proto = ct->tuplehash[IP_CT_DIR_ORIGINAL].tuple.dst.protonum;
	struct sk_buff *skb;

	skb = ctnetlink_event_build_msg(ct, ctinfo, proto, in, out);
	if (!skb)
		return;

	if (proto == IPPROTO_TCP) {
		nfnetlink_send(skb, 0, NFGRP_IPV4_CT_TCP, 0);
		return;
       	} else if (proto == IPPROTO_UDP) {
		nfnetlink_send(skb, 0, NFGRP_IPV4_CT_UDP, 0);
		return;
       	} else if (proto == IPPROTO_ICMP) {
		nfnetlink_send(skb, 0, NFGRP_IPV4_CT_ICMP, 0);
		return;
	} else {
		nfnetlink_send(skb, 0, NFGRP_IPV4_CT_OTHER, 0);
		return;
	}
	kfree_skb(skb);
	return;
}

#if 0
static void ctnetlink_destroy(struct ip_conntrack *ct)
{
        ctnetlink_create(ct, IP_CT_DELETE, NULL, NULL);
}
#endif

static inline int ctnetlink_kill(const struct ip_conntrack *i, void *data)
{
	struct ip_conntrack *t = (struct ip_conntrack *)data;

	if (!memcmp(&i->tuplehash[IP_CT_DIR_ORIGINAL], 
	    &t->tuplehash[IP_CT_DIR_ORIGINAL], 
	    sizeof(struct ip_conntrack_tuple_hash))) {
		//ip_conntrack_put(t);
		nf_conntrack_put(&t->infos[0]);
		return 1;
	}

	return 0;
}

static int
ctnetlink_del_conntrack(struct sock *ctnl, struct sk_buff *skb, 
			struct nlmsghdr *nlh, int *errp)
{
        struct ip_conntrack_tuple_hash *h;
	struct ip_conntrack_tuple *tuple;
	struct nfattr *cda[CTA_MAX];

	ct_debug(0, "entered\n");

	if (nfnetlink_check_attributes(ctnl_subsys, nlh, cda) < 0)
		return -EINVAL;

	if (cda[CTA_ORIG-1])
		tuple = NFA_DATA(cda[CTA_ORIG-1]);
	else {
		if (cda[CTA_RPLY-1])
			tuple = NFA_DATA(cda[CTA_RPLY-1]);
		else {
			ct_debug(0, "no tuple found in request\n");
			return -EINVAL;
		}
	}

	h = ip_conntrack_find_get(tuple, NULL);
        if (!h) {
		ct_debug(0, "tuple not found in conntrack hash:");
		DUMP_TUPLE(tuple);
		return -ENOENT;
	}

	ct_debug(0, "calling selective_cleanup\n");
	ip_ct_selective_cleanup(ctnetlink_kill, h->ctrack);

	return 0;
}

static int ctnetlink_done(struct netlink_callback *cb)
{
	ct_debug(0, "entering\n");
        return 0;
}

static int
ctnetlink_dump_build_msg(const struct ip_conntrack_tuple_hash *hash,
			 struct sk_buff *skb, u32 pid, u32 seq)
{
	struct ip_conntrack *ct;
	int err, proto;

	/* Only count originals */
	if (DIRECTION(hash))
		return 0;

	ct = hash->ctrack;
	if (!ct)
		goto nlmsg_failure;

	proto = ct->tuplehash[IP_CT_DIR_ORIGINAL].tuple.dst.protonum;
	err = ctnetlink_fill_info(skb, pid, seq, CTNL_MSG_NEWCONNTRACK, 1,
				  ct, NULL, proto, NULL, NULL);
	if (err <= 0)
		goto nlmsg_failure;
	return 0;

nlmsg_failure:
	if (skb)
		kfree_skb(skb);
	return -1;
}

static int
ctnetlink_dump_table(struct sk_buff *skb, struct netlink_callback *cb)
{
	int i;
	int idx;
	int s_idx = cb->args[0];

	/* Traverse hash; send originals then reply. */
	READ_LOCK(&ip_conntrack_lock);
        for (i = 0, idx = 0; i < ip_conntrack_htable_size; i++, idx++) {
		if (idx < s_idx)
			continue;
		if (LIST_FIND(&ip_conntrack_hash[i], ctnetlink_dump_build_msg,
			      struct ip_conntrack_tuple_hash *, skb,
			      NETLINK_CB(cb->skb).pid, cb->nlh->nlmsg_seq))
			continue;
	}
	READ_UNLOCK(&ip_conntrack_lock);

	cb->args[0] = idx;
	return skb->len;
}

static int
ctnetlink_get_conntrack(struct sock *ctnl, struct sk_buff *skb, 
			struct nlmsghdr *nlh, int *errp)
{
	struct ip_conntrack_tuple_hash *h;
	struct ip_conntrack_tuple *tuple;
	struct nfattr *cda[CTA_MAX];
	struct ip_conntrack *ct;
	struct sk_buff *skb2 = NULL;
	int err, proto;

	ct_debug(0, "entered\n");

	if (nlh->nlmsg_flags & NLM_F_DUMP) {
		struct nfgenmsg *msg = NLMSG_DATA(nlh);
		u32 rlen;

		if (msg->nfgen_family != AF_INET)
			return -EAFNOSUPPORT;

		ct_debug(0, "starting dump\n");
                if ((*errp = netlink_dump_start(ctnl, skb, nlh,
		    				ctnetlink_dump_table,
						ctnetlink_done)) != 0)
			return -EINVAL;
		rlen = NLMSG_ALIGN(nlh->nlmsg_len);
		if (rlen > skb->len)
			rlen = skb->len;
		skb_pull(skb, rlen);
		return 0;
	}

	if (nfnetlink_check_attributes(ctnl_subsys, nlh, cda) < 0)
		return -EINVAL;

	if (cda[CTA_ORIG-1])
		tuple = NFA_DATA(cda[CTA_ORIG-1]);
	else {
		if (cda[CTA_RPLY-1])
			tuple = NFA_DATA(cda[CTA_RPLY-1]);
		else
			return -EINVAL;
	}

	h = ip_conntrack_find_get(tuple, NULL);
	if (!h)
		return -ENOENT;

	ct = h->ctrack;
	if (!ct)
		goto nlmsg_failure;

	skb2 = alloc_skb(NLMSG_GOODSIZE, GFP_ATOMIC);
	if (!skb2)
		return -ENOMEM;
	NETLINK_CB(skb2).dst_pid = NETLINK_CB(skb).pid;

        proto = ct->tuplehash[IP_CT_DIR_ORIGINAL].tuple.dst.protonum;
	err = ctnetlink_fill_info(skb2, NETLINK_CB(skb).pid, nlh->nlmsg_seq, 
				  CTNL_MSG_NEWCONNTRACK, 1, ct, NULL, 
				  proto, NULL, NULL);
	ip_conntrack_put(ct);
	if (err <= 0)
		goto nlmsg_failure;

	err = netlink_unicast(ctnl, skb2, NETLINK_CB(skb).pid, MSG_DONTWAIT);
	if (err < 0)
		return err;
	return 0;

nlmsg_failure:
	if (skb2)
		kfree_skb(skb2);
	return -1;
}

/* Finish me: should support NLM_F_CREATE and NLM_F_REPLACE. */
static int 
ctnetlink_new_conntrack(struct sock *ctnl, struct sk_buff *skb, 
			struct nlmsghdr *nlh, int *errp)
{
	return -EOPNOTSUPP;
}


/* EXPECT */

static int
ctnetlink_exp_fill_info(struct sk_buff *skb, u32 pid, u32 seq,
		    int event, 
		    int nowait, 
		    const struct ip_conntrack_expect *exp)
{
	struct nlmsghdr *nlh;
	struct nfgenmsg *nfmsg;
	unsigned char *b;

	b = skb->tail;
	nlh = NLMSG_PUT(skb, pid, seq, (NFNL_SUBSYS_CTNETLINK<<8)|event, 
			sizeof(struct nfgenmsg));
	nlh->nlmsg_flags 	= (nowait && pid) ? NLM_F_MULTI : 0;
	nfmsg = NLMSG_DATA(nlh);
	nfmsg->nfgen_family = AF_INET;

	NFA_PUT(skb, CTA_EXP_TUPLE, sizeof(struct ip_conntrack_tuple),
		&exp->tuple);
	NFA_PUT(skb, CTA_EXP_MASK, sizeof(struct ip_conntrack_tuple),
		&exp->mask);
	NFA_PUT(skb, CTA_EXP_SEQNO, sizeof(u_int32_t), &exp->seq);
	NFA_PUT(skb, CTA_EXP_HELP, sizeof(union ip_conntrack_expect_help),
		&exp->help);

	/* FIXME: proto */

#ifdef CONFIG_IP_NF_NAT_NEEDED
#endif /* CONFIG_IP_NF_NAT_NEEDED */

	nlh->nlmsg_len = skb->tail - b;
	return skb->len;

nlmsg_failure:
nfattr_failure:
	skb_trim(skb, b - skb->data);
	return -1;
}

static inline struct sk_buff *
ctnetlink_exp_event_build_msg(const struct ip_conntrack_expect *exp)
{
	struct sk_buff *skb;
	int err;

	skb = alloc_skb(NLMSG_GOODSIZE, GFP_ATOMIC);
	if (!skb)
		return NULL;

	err = ctnetlink_exp_fill_info(skb, 0, 0, CTNL_MSG_NEWEXPECT, 1, exp);
	if (err <= 0)
		goto nlmsg_failure;
	return skb;

nlmsg_failure:
	if (skb)
		kfree_skb(skb);
        return NULL;
}

static void
ctnetlink_exp_create(struct ip_conntrack_expect *exp)
{
	u16 proto = exp->tuple.dst.protonum;
	struct sk_buff *skb;

	skb = ctnetlink_exp_event_build_msg(exp);
	if (!skb)
		return;

	if (proto == IPPROTO_TCP) {
		nfnetlink_send(skb, 0, NFGRP_IPV4_CT_TCP, 0);
		return;
       	} else if (proto == IPPROTO_UDP) {
		nfnetlink_send(skb, 0, NFGRP_IPV4_CT_UDP, 0);
		return;
       	} else if (proto == IPPROTO_ICMP) {
		nfnetlink_send(skb, 0, NFGRP_IPV4_CT_ICMP, 0);
		return;
	} else {
		nfnetlink_send(skb, 0, NFGRP_IPV4_CT_OTHER, 0);
		return;
	}
	kfree_skb(skb);
	return;
}


static int
ctnetlink_del_expect(struct sock *ctnl, struct sk_buff *skb, 
		     struct nlmsghdr *nlh, int *errp)
{
        struct ip_conntrack_expect *exp;
	struct ip_conntrack_tuple *tuple;
	struct nfattr *cda[CTA_MAX];

	if (nfnetlink_check_attributes(ctnl_subsys, nlh, cda) < 0)
		return -EINVAL;

	if (cda[CTA_ORIG-1])
		tuple = NFA_DATA(cda[CTA_ORIG-1]);
	else {
		if (cda[CTA_RPLY-1])
			tuple = NFA_DATA(cda[CTA_RPLY-1]);
		else
			return -EINVAL;
	}

	/* bump usage count to 2 */
	exp = ip_conntrack_expect_find_get(tuple);
	if (!exp)
		return -ENOENT;

	/* after list removal, usage count == 1 */
	ip_conntrack_unexpect_related(exp);
	/* we have put what we 'get' above. after this line usage count == 0 */
	ip_conntrack_expect_put(exp);

	return 0;
}

static int
ctnetlink_exp_dump_build_msg(const struct ip_conntrack_expect *exp,
			 struct sk_buff *skb, u32 pid, u32 seq)
{
	int err, proto;

	proto = exp->tuple.dst.protonum;
	err = ctnetlink_exp_fill_info(skb, pid, seq, CTNL_MSG_NEWEXPECT, 1, 
				      exp);
	if (err <= 0)
		goto nlmsg_failure;
	return 0;

nlmsg_failure:
	if (skb)
		kfree_skb(skb);
	return -1;
}

static int
ctnetlink_exp_dump_table(struct sk_buff *skb, struct netlink_callback *cb)
{
	ct_debug(0, "entered\n");
	if (cb->args[0] == 0) {
		READ_LOCK(&ip_conntrack_lock);
		LIST_FIND(&ip_conntrack_expect_list, 
			  ctnetlink_exp_dump_build_msg,
			  struct ip_conntrack_expect *, skb,
			  NETLINK_CB(cb->skb).pid, cb->nlh->nlmsg_seq);
		READ_UNLOCK(&ip_conntrack_lock);
		cb->args[0] = 1;
	}
	ct_debug(0, "returning\n");

	return skb->len;
}


static int
ctnetlink_get_expect(struct sock *ctnl, struct sk_buff *skb, 
		     struct nlmsghdr *nlh, int *errp)
{
	struct ip_conntrack_expect *exp;
	struct ip_conntrack_tuple *tuple;
	struct nfattr *cda[CTA_MAX];
	struct sk_buff *skb2 = NULL;
	int err, proto;

	ct_debug(0, "entered\n");

	if (nlh->nlmsg_flags & NLM_F_DUMP) {
		struct nfgenmsg *msg = NLMSG_DATA(nlh);
		u32 rlen;

		if (msg->nfgen_family != AF_INET)
			return -EAFNOSUPPORT;

		ct_debug(0, "starting dump\n");
                if ((*errp = netlink_dump_start(ctnl, skb, nlh,
		    				ctnetlink_exp_dump_table,
						ctnetlink_done)) != 0)
			return -EINVAL;
		rlen = NLMSG_ALIGN(nlh->nlmsg_len);
		if (rlen > skb->len)
			rlen = skb->len;
		skb_pull(skb, rlen);
		return 0;
	}

	if (nfnetlink_check_attributes(ctnl_subsys, nlh, cda) < 0)
		return -EINVAL;

	if (cda[CTA_ORIG-1])
		tuple = NFA_DATA(cda[CTA_ORIG-1]);
	else {
		if (cda[CTA_RPLY-1])
			tuple = NFA_DATA(cda[CTA_RPLY-1]);
		else
			return -EINVAL;
	}

	exp = ip_conntrack_expect_find_get(tuple);
	if (!exp)
		return -ENOENT;

	skb2 = alloc_skb(NLMSG_GOODSIZE, GFP_ATOMIC);
	if (!skb2)
		return -ENOMEM;
	NETLINK_CB(skb2).dst_pid = NETLINK_CB(skb).pid;
	proto = exp->tuple.dst.protonum;
	
	err = ctnetlink_exp_fill_info(skb2, NETLINK_CB(skb).pid, 
				      nlh->nlmsg_seq, CTNL_MSG_NEWEXPECT,
				      1, exp);
	if (err <= 0)
		goto nlmsg_failure;

	err = netlink_unicast(ctnl, skb2, NETLINK_CB(skb).pid, MSG_DONTWAIT);
	if (err < 0)
		return err;
	return 0;

nlmsg_failure:
	if (skb2)
		kfree_skb(skb2);
	return -1;
}

static int
ctnetlink_new_expect(struct sock *ctnl, struct sk_buff *skb,
		     struct nlmsghdr *nlh, int *errp)
{
	return -EOPNOTSUPP;
}

/* struct conntrack_expect stuff */

#if 0
static struct ip_conntrack_notify ctnl_notify = { { NULL, NULL },
						ctnetlink_destroy,
						ctnetlink_create };

static struct ip_conntrack_notify ctnl_exp_notify;
#endif

static void __exit ctnetlink_exit(void)
{
	printk("ctnetlink: unregistering with nfnetlink.\n");
//	ip_conntrack_notify_unregister(&ctnl_exp_notify);
//	ip_conntrack_notify_unregister(&ctnl_notify);
	nfnetlink_subsys_unregister(ctnl_subsys);
	kfree(ctnl_subsys);
	return;
}

static int __init ctnetlink_init(void)
{
	int ret;

	ctnl_subsys = nfnetlink_subsys_alloc(CTNL_MSG_COUNT);
	if (!ctnl_subsys) {
		ret = -ENOMEM;
		goto err_out; 
	}

	ctnl_subsys->name = "conntrack";
	ctnl_subsys->subsys_id = NFNL_SUBSYS_CTNETLINK;
	ctnl_subsys->cb_count = CTNL_MSG_COUNT;
	ctnl_subsys->attr_count = CTA_MAX;
	ctnl_subsys->cb[CTNL_MSG_NEWCONNTRACK].call = ctnetlink_new_conntrack;
	ctnl_subsys->cb[CTNL_MSG_NEWCONNTRACK].cap_required = CAP_NET_ADMIN;
	ctnl_subsys->cb[CTNL_MSG_DELCONNTRACK].call = ctnetlink_del_conntrack;
	ctnl_subsys->cb[CTNL_MSG_DELCONNTRACK].cap_required = CAP_NET_ADMIN;
	ctnl_subsys->cb[CTNL_MSG_GETCONNTRACK].call = ctnetlink_get_conntrack;
	ctnl_subsys->cb[CTNL_MSG_GETCONNTRACK].cap_required = 0;
	ctnl_subsys->cb[CTNL_MSG_NEWEXPECT].call = ctnetlink_new_expect;
	ctnl_subsys->cb[CTNL_MSG_NEWEXPECT].cap_required = CAP_NET_ADMIN;
	ctnl_subsys->cb[CTNL_MSG_DELEXPECT].call = ctnetlink_del_expect;
	ctnl_subsys->cb[CTNL_MSG_DELEXPECT].cap_required = CAP_NET_ADMIN;
	ctnl_subsys->cb[CTNL_MSG_GETEXPECT].call = ctnetlink_get_expect;
	ctnl_subsys->cb[CTNL_MSG_GETEXPECT].cap_required = 0;
	// FIXME: CONFIRM

	printk("ctnetlink: registering with nfnetlink v%s.\n", ctversion);
	if (nfnetlink_subsys_register(ctnl_subsys) < 0) {
		printk("ctnetlink_init: cannot register with nfnetlink.\n");
		ret = -1;
		goto err_free_subsys;
	}


#if 0
	if (ip_conntrack_notify_register(&ctnl_notify) < 0) {
		printk("ctnetlink_init: cannot register notifier.\n");
		ret = -1;
		goto err_unreg_subsys;
	}

	if (ip_conntrack_notify_register(&ctnl_exp_notify) < 0) {
		printk("ctnetlink_init: cannot register exp notifier\n");
		ret = -1;
		goto err_unreg_notify;
	}
#endif


	return 0;
	
#if 0
err_unreg_notify:
	ip_conntrack_notify_unregister(&ctnl_notify);
#endif 
err_unreg_subsys:
	nfnetlink_subsys_unregister(ctnl_subsys);
err_free_subsys:
	kfree(ctnl_subsys);
err_out:
	return ret;
}

module_init(ctnetlink_init);
module_exit(ctnetlink_exit);
